/* sqfigb.c -- Square-Figure (B version [binary, bitmap, plan b, ...])
** functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqfigb.c,v 1.3 2003/10/05 10:07:34 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqfigb.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "sqfigbut.h"


#if !defined(OPTIM)
#define OPTIM 9
#endif


/** check for connectivity */
static int connectivityCheck(int const* bitmap, int ySize, int xSize)
{
	int y, x;
	int b[32], c[32], cy;
	int differs;

	/* [unused parameters] */
	(void)xSize;

#if OPTIM < 2
	if (ySize == 0 || xSize == 0 || bitmap == NULL) {
		return 1;
	}
#endif

	/* find first pixel */
	firstBitVector(bitmap, ySize, &y, &x);

	if (y == -1) {

#ifdef DEBUG
		printf("rejected for empty bitmap\n");
#endif

		return 1;
	}

	/* initialize */
	initVector(
		b,

#if OPTIM < 2
		sizeof(b) / sizeof(*b)
#else
		ySize
#endif

	);
	initVector(
		c,

#if OPTIM < 2
		sizeof(c) / sizeof(*c)
#else
		ySize
#endif

	);
	/* start with first found pixel */
	b[y] |= BIT(x);
	c[y] = b[y];

	do {

		/* propagate pixels */
		for (y = 0; y < ySize; y++) {
			cy = c[y];

			/* propagate up */
			if (y < ySize - 1) {
				cy |= b[y + 1];
			}
			/* propagate down */
			if (y > 0) {
				cy |= b[y - 1];
			}
			/* propagate left */
			cy |= (b[y] >> 1);
			/* propagate right */
			cy |= (b[y] << 1);

			/* reduce to bitmap */
			cy &= bitmap[y];

			c[y] = cy;
		}

		/* check for progress; copy back */
		differs = 0;
		for (y = 0; y < ySize; y++) {
			if (c[y] != b[y]) {
				differs = 1;
				b[y] = c[y];
			}
		}

	} while (differs);

	/* check with bitmap */
	for (y = 0; y < ySize; y++) {
		if (c[y] != bitmap[y]) {

#ifdef DEBUG2
			printBitmapCallback(bitmap, ySize, xSize, NULL);
#endif

#ifdef DEBUG
			printf("rejected for missing bitmap connectivity\n");
#endif

			return 0;
		}
	}

#ifdef DEBUG
	printf("* passed\n");
#endif

	return 1;
}


#if OPTIM >= 1
/** Optimization hints to reduce the comb(ySize*xSize,pixels) possible
* bitmaps to a subset containing all the figure solutions (that are
* later (possibly) further reduced by the filter callback).
* Return values: 0 -- ok, 1 -- bad (skip), 2 -- beyond (don't continue
* in this layer).
*/
static int checkHints(
	int* bitmap,
	int ySize,
	int xSize,
	int y,
	int x,
	int pixels)
{

#if OPTIM >= 3
	int by, c;
#endif

	/* [unused parameters] */
	(void)ySize;

#ifdef DEBUG
	/* note: only painting up to y (not ySize),
	* whole xSize is used though
	*/
	printBitmapCallback(bitmap, y + 1, xSize, NULL);
#endif

	/* tests to be done if line is completed */
	if (x == xSize - 1) {

		/* the first row must contain pixels for the figure to be top aligned */
		/* subsequent rows must contain pixels in order to be connected */
		/* [this assumes trailing unused bits must be 0] */
		/* note: based on the sequence of generated bitmaps, this test may not trigger */
		if (bitmap[y] == 0) {

#ifdef DEBUG
			if (y == 0) {
				printf("rejected for empty line (missing top alignment) (detected at end of line)\n");
			} else {
				printf("rejected for empty line (missing line connectivity) (detected at end of line)\n");
			}
#endif

			return 1;
		}

		/* every layer must be connected to the above layer */
		/* that means: there must be at least one bit set in both ints */
		if (y > 0) {
			if ((bitmap[y] & bitmap[y - 1]) == 0) {

#ifdef DEBUG
				printf("rejected for missing line connectivity (detected at end of line)\n");
#endif

				return 1;
			}
		}
	}

	/* tests to be done at the start of a subsequent line */
	/* these are essentially the same tests as above */
	/* these tests return 2 to indicate the previous pixel has to be moved */
	if (x == 0 && y > 0) {

		/* the first row must contain pixels for the figure to be top aligned */
		/* subsequent rows must contain pixels in order to be connected */
		if (bitmap[y - 1] == 0) {

#ifdef DEBUG
			if (y == 1) {
				printf("rejected for empty line (missing top alignment) (detected at beginning of line)\n");
			} else {
				printf("rejected for empty line (missing line connectivity) (detected at beginning of line)\n");
			}
#endif

			return 2;
		}

		/* every layer must be connected to the above layer */
		/* that means: there must be at least one bit set in both ints */
		if (y > 1) {
			if ((bitmap[y - 1] & bitmap[y - 2]) == 0) {

#ifdef DEBUG
				printf("rejected for missing line connectivity (detected at beginning of line)\n");
#endif

				return 2;
			}
		}
	}

	/* tests to be done if the number of pixels is exhausted (which means that the
	* line is completed too (but the start-of-line tests have to be done before))
	*/
	/* x may be 0 (start of line) */
	if (pixels == 1 && x != xSize - 1) {

		/* the first row must contain pixels for the figure to be top aligned */
		/* subsequent rows must contain pixels in order to be connected */
		/* note: based on the sequence of generated bitmaps, this test may not trigger */
		if (bitmap[y] == 0) {

#ifdef DEBUG
			if (y == 0) {
				printf("rejected for empty line (missing top alignment) (detected at end of pixels)\n");
			} else {
				printf("rejected for empty line (missing line connectivity) (detected at end of pixels)\n");
			}
#endif

			return 1;
		}

		/* every layer must be connected to the above layer */
		/* that means: there must be at least one bit set in both ints */
		if (y > 0) {
			if ((bitmap[y] & bitmap[y - 1]) == 0) {

#ifdef DEBUG
				printf("rejected for missing line connectivity (detected at end of pixels)\n");
#endif

				return 1;
			}
		}
	}

	/* tests to be done on the last line */
	if (y == ySize - 1) {

		/* check for left alignment */
		if (!bitSetVectorP(bitmap, ySize, 0)) {

#ifdef DEBUG
			printf("rejected for missing left alignment (detected at last line)\n");
#endif

			return 1;
		}
	}

	/* tests to be done if the number of pixels is exhausted (which means that this
	* is the last line too)
	*/
	/* these tests return 2 too to indicate the previous pixel has to be moved */
	if (pixels == 1 && y != ySize - 1) {

		/* check for left alignment */
		if (!bitSetVectorP(bitmap, y + 1, 0)) {

#ifdef DEBUG
			printf("rejected for missing left alignment (detected at end of pixels)\n");
#endif

			return 2;
		}
	}

#if OPTIM >= 3
	if (y < ySize - 2) {

		/* check stray pixels */

		/* save state */
		by = bitmap[y];

		/* fill rest of line */
		bitmap[y] |= BITSFROMTO(x + 1, xSize);
		/* fill next line */
		bitmap[y + 1] = BITSTO(xSize);

		c = connectivityCheck(bitmap, y + 2, xSize);

		/* reset state */
		bitmap[y] = by;
		bitmap[y + 1] = 0;

		if (!c) {
			return 1;
		}
	}
#endif

	return 0;
}
#endif


/** Engine to generate the bitmaps.  The optimized hints already reduce
* the comb(ySize*xSize,pixels) possible bitmaps to a subset containing
* all the figure solutions (that are (possibly) further reduced by the
* filter callback), which makes the function name less accurate.  The
* function is recursive.
*/
static void generateSubBitmaps(
	int* bitmap,
	int ySize,
	int xSize,
	int yFrom, /* 0..ySize [including ySize!] */
	int xFrom, /* 0..xSize [including xSize!] */
	int pixels,
	void (*callback)(int const* bitmap, int ySize, int xSize, void* userData),
	void* userData)
{
	int xFromEffective, y, x, by;

#if OPTIM >= 1
	int hint;
#endif

#if OPTIM < 2
	/* [sanity checks] */
	if (pixels < 0) {
		return;
	}
#endif

	/* recursion termination */
	if (pixels == 0) {
		if (callback != NULL) {
			(*callback)(bitmap, ySize, xSize, userData);
		}

		return;
	}

#if OPTIM < 2
	/* [sanity checks] */
	if (bitmap == NULL || xSize > 8 * sizeof(int)) {
		return;
	}
#endif

	xFromEffective = xFrom;
	for (y = yFrom; y < ySize; y++) {
		for (x = xFromEffective; x < xSize; x++) {

			/* save state */
			by = bitmap[y];

			bitmap[y] |= BIT(x);

#if OPTIM >= 1
			hint = checkHints(bitmap, ySize, xSize, y, x, pixels);

			if (hint == 1) { /* bad (skip) */
				bitmap[y] = by;
				continue;
			}

			if (hint == 2) { /* beyond (don't continue in this layer) */
				bitmap[y] = by;
				return;
			}
#endif

			/* recursion */
			generateSubBitmaps(
				bitmap,
				ySize,
				xSize,
				y, /* yFrom */
				x + 1, /* xFrom */
				pixels - 1, /* pixels */
				callback,
				userData);

			/* restore state */
			bitmap[y] = by;
		}

		xFromEffective = 0;
	}
}


static void generateBitmaps(
	int* bitmap,
	int ySize,
	int xSize,
	int pixels,
	void (*callback)(int const* bitmap, int ySize, int xSize, void* userData),
	void* userData)
{

#if OPTIM < 2
	/* [sanity checks] */
	if (bitmap == NULL && ySize > 0 && xSize > 0) {
		return;
	}
#endif

	if (xSize > 0) {
		initVector(bitmap, ySize);
	}

	generateSubBitmaps(
		bitmap,
		ySize,
		xSize,
		0, /* yFrom */
		0, /* xFrom */
		pixels,
		callback,
		userData);
}


/** Filter callback to reduce the solutions to the figures (by assuring
* connectivity and alignment, etc).  Note that some of this reduction
* may have been done by the hints earlier (as an optimization).
*/
static void figureFilterCallback(
	int const* bitmap,
	int ySize,
	int xSize,
	void* userData)
{
	void (*nextCallback)(int const* bitmap, int ySize, int xSize, void* userData);
	void* nextUserData;

	if (ySize == 0 || xSize == 0) {
		/* note: this is _not_ a sanity check */
		/* does not count as solution */
		return;
	}

#if OPTIM < 2
	/* [sanity checks] */
	if (bitmap == NULL) {
		return;
	}
#endif

	/* check for top alignment */
	/* this assumes trailing unused bits must be 0 */
	/* based on OPTIM level, this has already been checked earlier */
	if (bitmap[0] == 0) {

#ifdef DEBUG
		printf("rejected for missing top alignment\n");
#endif

		return;
	}

	/* check for left alignment */
	/* based on OPTIM level, this has already been checked earlier */
	if (!bitSetVectorP(bitmap, ySize, 0)) {

#ifdef DEBUG
		printf("rejected for missing left alignment\n");
#endif

		return;
	}

	/* check for connectivity */
	if (!connectivityCheck(bitmap, ySize, xSize)) {
		return;
	}

	nextCallback = *(void (**)(int const*, int, int, void*))((void**)userData)[0];
	nextUserData = *(void**)((void**)userData)[1];
	(*nextCallback)(bitmap, ySize, xSize, nextUserData);
}


/** Generate the figures and call a call-back function for each one.
*/
void generateFigures(
	int size,
	void (*callback)(int const* bitmap, int ySize, int xSize, void* userData),
	void* userData)
{
	int bitmap[32];
	void* data[2];

#if OPTIM < 2
	/* [sanity checks] */
	if (size < 0 || size > 32) {
		return;
	}
#endif

	/* initialize to zero; filters may require this */
	initVector(
		bitmap,

#if OPTIM < 2
		sizeof(bitmap) / sizeof(*bitmap)
#else
		size
#endif

	);

	/* [additional indirection: work around compiler problem casting from/to text/data] */
	data[0] = (void*)&callback;
	data[1] = (void*)&userData;
	generateBitmaps(
		bitmap,
		size, /* ySize */
		size, /* xSize */
		size, /* pixels */
		figureFilterCallback,
		data);
}


/** test whether this bitmap is the primary one, i.e. does not have
* rotameres or mirrors that come earlier if ordered
*/
static int testPrimaryBitmap(int const* bitmap, int ySize, int xSize)
{
	int y, x;
	int i, m;

#if OPTIM < 2
	if (bitmap == NULL) {
		return 1;
	}
#endif

	adjustDimensions(bitmap, &ySize, &xSize);

	/* X must not be less than Y,
	* (that one could be rotated up)
	*/
	if (ySize > xSize) {
		return 0;
	}

	/* symmetry bit vector (i.e. symmetries to check) */
	m = ~0;

	/* (go over all pixels, though half of them would be sufficient
	* (but not always the same half))
	*/
	for (y = 0; y < ySize; y++) {
		for (x = 0; x < xSize; x++) {

			i = 0;

/* helper macro; references local variables and does program flow */
#define TESTSYMMETRY(yy, xx) \
	{ \
		if (BITSETP(m, i)) { \
			switch (BITSETP(bitmap[y], x) - BITSETP(bitmap[(yy)], (xx))) { \
			case 1 : \
				/* we're primary relative to this symmetry; \
				* don't need to check this symmetry anymore \
				*/ \
				m &= ~BIT(i); \
				break; \
			case -1 : \
				/* not primary */ \
				return 0; \
			} \
		} \
		i++; \
	}

			/* compare top to bottom */
			TESTSYMMETRY((ySize - 1) - y, x);

			/* compare left to right */
			TESTSYMMETRY(y, (xSize - 1) - x);

			/* compare top left to bottom right */
			TESTSYMMETRY((ySize - 1) - y, (xSize - 1) - x);

			if (ySize != xSize) {
				/* don't need to check diagonal mirrorring unless ySize==xSize */
				continue;
			}

			/* additionally invert the dimensions */

			TESTSYMMETRY(x, y);

			TESTSYMMETRY(x, (ySize - 1) - y);

			TESTSYMMETRY((xSize - 1) - x, y);

			TESTSYMMETRY((xSize - 1) - x, (ySize - 1) - y);

			/* check if we're done */
			if (firstBit(m) == i) {
				return 1;
			}
		}
	}

	/* we're not non-primary and have at least one symmetry */

	return 1;
}


static void countCallback(int const* bitmap, int ySize, int xSize, void* userData)
{
	int *count, *distinctCount;

	/* [unused parameters] */
	(void)bitmap;
	(void)ySize;
	(void)xSize;

	count = (int*)((void**)userData)[0];
	distinctCount = (int*)((void**)userData)[1];

	if (count != NULL) {
		(*count)++;
	}

	if (distinctCount != NULL) {
		if (testPrimaryBitmap(bitmap, ySize, xSize)) {
			(*distinctCount)++;
		}
	}
}


/** Count the figures.
*/
void countFigures(int size, int* count, int* distinctCount)
{
	void* data[2];

	if (count != NULL) {
		*count = 0;
	}
	if (distinctCount != NULL) {
		*distinctCount = 0;
	}

	if (count != NULL || distinctCount != NULL) {
		data[0] = (void*)count;
		data[1] = (void*)distinctCount;
		generateFigures(
			size,
			countCallback,
			(void*)data);
	}
}


static void printBitmap(int const* bitmap, int ySize, int xSize)
{
	int y, x;

	/* note: one could reduce the dimensions to those actually
	* containing pixels, which we don't
	* i.e.: it's left to the caller to do so
	*/

	/* (one) (upper) frame */
	for (x = 0; x < xSize; x++) {
		printf("-");
	}
	printf("\n");

	for (y = 0; y < ySize; y++) {
		for (x = 0; x < xSize; x++) {
			printf("%s", (BITSETP(bitmap[y], x) ? "#" : "."));
		}
		printf("\n");
	}
}


/** Print a bitmap.
*/
void printBitmapCallback(int const* bitmap, int ySize, int xSize, void* userData)
{
	/* [unused parameters] */
	(void)userData;

	printBitmap(bitmap, ySize, xSize);
}

