/* sqfigbut.c -- Square-Figure (B version) utilities.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqfigbut.c,v 1.2 2003/10/05 10:07:12 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqfigbut.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#if !defined(OPTIM)
#define OPTIM 9
#endif


/** Check for any nth bit in a vector.
*/
int bitSetVectorP(int const* vector, int size, int n)
{
	int i;

#if OPTIM < 2
	/* [sanity checks] */
	if (vector == NULL || size < 0) {
		return 0;
	}
#endif

	for (i = 0; i < size; i++) {
		if (BITSETP(vector[i], n)) {
			return 1;
		}
	}

	return 0;
}


/** Initialize a vector.
*/
void initVector(int* vector, int size)
{
	int i;

#if OPTIM < 2
	/* [sanity checks] */
	if (vector == NULL || size < 0) {
		return;
	}
#endif

	for (i = 0; i < size; i++) {
		vector[i] = 0;
	}
}


/** Highest populated bit.
*/
static int hpb(unsigned int n)
{

#if OPTIM >= 4 && defined(OPTIM32)

	/* fast version (for 32 bit architectures) */

	int i;
	unsigned int p;

	/*assert(sizeof(n) == 4);*/

	if (n == 0) {
		return -1;
	}

	i = 0;

/* helper macro; references local variables
* [does though single evaluation]
*/
#define BITTEST(m) \
	i <<= 1; \
	p = n & (m); \
	if (p != 0) { \
		i |= 1; \
		n = p; \
	}

	BITTEST(0xffff0000u);
	BITTEST(0xff00ff00u);
	BITTEST(0xf0f0f0f0u);
	BITTEST(0xccccccccu);
	BITTEST(0xaaaaaaaau);

	return i;

#else

	/* slow version */

	int i;

#if OPTIM >= 2
	if (n == 0) {
		return -1;
	}
#endif

	/* note: (slow) linear search */

	for (i = MAXBITS - 1; i >= 0; i--) {
		if (BITSETP(n, i)) {
			return i;
		}
	}

	return -1;

#endif

}


/** Lowest populated bit.
*/
static int lpb(unsigned int n)
{

#if OPTIM >= 4 && defined(OPTIM32)

	/* fast version (for 32 bit architectures) */

	int i;
	unsigned int p;

	/*assert(sizeof(n) == 4);*/

	if (n == 0) {
		return -1;
	}

	i = 0;

	/* [helper macro BITTEST is the same as in hpb] */

	BITTEST(~0xffff0000u);
	BITTEST(~0xff00ff00u);
	BITTEST(~0xf0f0f0f0u);
	BITTEST(~0xccccccccu);
	BITTEST(~0xaaaaaaaau);

	return 31 - i;

#else

	/* slow version */

	int i;

#if OPTIM >= 2
	if (n == 0) {
		return -1;
	}
#endif

	/* note: (slow) linear search */

	for (i = 0; i < MAXBITS; i++) {
		if (BITSETP(n, i)) {
			return i;
		}
	}

	return -1;

#endif

}


/** Find first bit.
* Returns -1 if none is found.
*/
int firstBit(int i)
{
	return lpb((unsigned int)i);
}


/** Find first bit in a vector.
*/
void firstBitVector(int const* vector, int ySize, int* y, int* x)
{
	int yy, xx;

	if (y != NULL) {
		*y = -1;
	}
	if (x != NULL) {
		*x = -1;
	}

#if OPTIM < 2
	/* [sanity checks] */
	if (vector == NULL || ySize < 0) {
		return;
	}
#endif

	for (yy = 0; yy < ySize; yy++) {
		xx = firstBit(vector[yy]);
		if (xx != -1) {
			if (y != NULL) {
				*y = yy;
			}
			if (x != NULL) {
				*x = xx;
			}
			return;
		}
	}
}


/** Find last bit.
* Returns -1 if none is found.
*/
static int lastBit(int i)
{
	return hpb((unsigned int)i);
}


static int vectorOr(int const* vector, int size)
{
	int orValue;
	int i;

#if OPTIM < 2
	/* [sanity checks] */
	if (vector == NULL || size < 0) {
		size = 0;
	}
#endif

	orValue = 0;
	for (i = 0; i < size; i++) {
		orValue |= vector[i];
	}

	return orValue;
}


/** Adjust dimensions.
*/
void adjustDimensions(int const* bitmap, int* ySize, int* xSize)
{

#if OPTIM < 2
	/* [sanity checks] */
	if (bitmap == NULL || ySize == NULL || xSize == NULL) {
		return;
	}
#endif

	if (*ySize <= 0 || *xSize <= 0) {
		*ySize = *xSize = 0;
		return;
	}

	while (*ySize > 0 && bitmap[*ySize - 1] == 0) {
		(*ySize)--;
	}

	if (*ySize == 0) {
		*xSize = 0;
		return;
	}

	*xSize = lastBit(vectorOr(bitmap, *ySize)) + 1;

	if (*xSize == 0) {
		*ySize = 0;
		return;
	}
}


